(* Time-stamp: <modified the 06/02/2017 (at 11:11) by jahier> *)
open Soc
open Data
open SocExecValue

(* exported *)
let rec (do_step : Soc.tbl -> Soc.t -> SocExecValue.ctx -> SocExecValue.ctx) =
  fun soc_tbl soc ctx ->
    let step = match soc.step with [step] -> step | _ ->  assert false in
    let ctx = SocExec.soc_step step soc_tbl soc ctx in
    ctx

(****************************************************************************)
(* XXX duplication of SocExec ! 

Maybe this code below, that is more general, is as (in)efficient ?

*)
let dbg = Lv6Verbose.get_flag "execDbg"
let profile_info = Lv6Verbose.profile_info

let (assign_expr : SocExecValue.ctx -> var_expr -> var_expr -> SocExecValue.ctx) =
  fun ctx ve_in ve_out -> (* ve_out := ve_in (in ctx) *)
  Lv6Verbose.exe
    ~flag:dbg 
    (fun () ->
     print_string ("\nAssigning "^(SocUtils.string_of_filter ve_out) ^
                     " to " ^(SocUtils.string_of_filter ve_in) ^ "... ");
      flush stdout);
    let res =
    { ctx with s =
        let v = SocExecValue.get_value ctx ve_in in
        sadd_partial ctx.s ve_out ctx.cpath v
    }
    in
    Lv6Verbose.exe ~flag:dbg (fun () -> print_string (" Done!"); flush stdout);
    res
      
(* [array_index i v] returns the var_expr v[i] *)
let (array_index : int -> var -> var_expr) =
  fun i (vn,vt) -> 
    let vt_elt = 
      match vt with
        | Array(vt_elt,_) -> vt_elt
        | _ -> assert false
    in
    Index(Var(vn,vt),i,vt_elt)

let get_all_subst = SocExecValue.filter_top_subst
let get_all_subst = SocExecValue.substs_to_data_subst
let get_input_vals val_ctx in_vars =
  let datal = SocExecValue.get_vals val_ctx in
  let datal = List.filter (fun (x,_) -> List.mem_assoc x in_vars) datal in
  datal
  
exception  SourceInfoError of (string * string)
    
  
let rec (soc_step :  Lxm.t -> Soc.step_method -> Soc.tbl -> Soc.t ->
         Event.t -> SocExecValue.ctx ->
         (Event.t -> SocExecValue.ctx -> Event.t) -> Event.t) =
  fun lxm step soc_tbl soc ectx val_ctx cont ->
    profile_info ("SocExecDbg.soc_step \n");
    let soc_name,_,_ = soc.key in
    let event = 
      match step.impl with
      | Extern ->
        print_string (
          "\nextern nodes and functions not yet supported in the interpreter, sorry.\n"^
            "Please use the C code generator (-2c)."
        );
        exit 2
      | Predef -> (
        try 
          let val_ctx = SocExecEvalPredef.get soc.key val_ctx in
          cont ectx val_ctx 
        with Not_found -> (* Not a predef op *) print_string (
          "*** internal error in "^soc_name^". Is it defined in SocExecEvalPredef?\n");
          flush stdout; assert false
      )
      | Gaol(vl,gaol) -> do_gaol soc_tbl ectx gaol val_ctx cont
        
      | Boolred(i,j,k) -> (
        (* XXX mettre ce code dans socPredef ? (ou socMetaopPredef)*)
        let inputs, outputs = soc.profile in
        let b_array = (List.hd inputs) in
        let cpt = ref 0 in
        for i = 0 to k-1 do
          let a_i = array_index i b_array in
          let v = SocExecValue.get_value val_ctx a_i in
          if v = B true then incr cpt;
        done;
        let res = B (!cpt >= i && !cpt <= j) in
        let res_var = fst (List.hd outputs) in
        let s = sadd val_ctx.s (res_var::val_ctx.cpath) res in
        cont ectx { val_ctx with s = s }
      )
      | Condact(node_sk, dft_cst) -> (
        let clk = SocExecValue.get_value val_ctx (Var ("activate",Bool)) in
        let vel_in, vel_out = soc.profile in
        let vel_in  =  List.map (fun x -> Var x) (List.tl vel_in) in
        let vel_out =  List.map (fun x -> Var x) vel_out in
        let node_soc = SocUtils.find lxm node_sk soc_tbl in
        let inst_name =
          match soc.instances with
          | [] -> let (proc_name,_,_) = node_soc.key in proc_name
          | [inst] -> fst inst
          | _ -> assert false
        in
        let path_saved = val_ctx.cpath in
        let val_ctx = { val_ctx with cpath=inst_name::val_ctx.cpath } in
        if clk = B true then (
          let node_step = match node_soc.step with
              [step] -> step
            | _ -> assert false
          in
          let cont ectx val_ctx =
            let val_ctx =
              { 
                cpath=path_saved;
                s = sadd val_ctx.s ("_memory"::val_ctx.cpath) (B false)
              }
            in
            cont ectx val_ctx
          in
          do_soc_step lxm inst_name node_step val_ctx soc_tbl node_soc
            vel_in vel_out ectx cont              
        ) else (
          let first_step = Var ("_memory",Bool) in
          let val_ctx =  { val_ctx with cpath=path_saved } in
          let v = get_value val_ctx first_step in
          let val_ctx =
            if v = U || v = B true then
              (* We are on the first step of node_soc;
                 - we assign the output var to the default values *)
              (assert (List.length dft_cst = List.length vel_out);
               List.fold_left2 assign_expr val_ctx dft_cst vel_out)
            else
              (* We are not on the first step of node_soc; hence we do nothing 
                 and the output will keep their previous value. *)
              val_ctx
          in
          let val_ctx = { val_ctx with
            s = sadd val_ctx.s ("_memory"::val_ctx.cpath) (B false) }
          in
          cont ectx val_ctx
        )
      )
      | Iterator(iter, node_sk, n) -> 
        let node_soc = SocUtils.find lxm node_sk soc_tbl in
        let node_step = match node_soc.step with [step] -> step | _ ->  assert false in
        let iter_inputs,iter_outputs = soc.profile in
        let (proc_name,_,_) = node_soc.key in 
        let inst_name =
          match soc.instances with
          | [] -> Array.make n proc_name
          | _  -> Array.of_list (List.map fst soc.instances)
        in
        (* As we need to iterate on an (instance) array, we locally switch
           to the evildark side *)
        let rval_ctx = ref val_ctx in
        let ref_event = ref ectx (* XXX dummy ?? *) in
        for i = 0 to n-1 do
          rval_ctx := { !rval_ctx with cpath = inst_name.(i)::val_ctx.cpath };
          let vel_in, vel_out =
            match iter with
            | "map" -> (List.map (array_index i) iter_inputs,
                        List.map (array_index i) iter_outputs)
            | "fold" | "red" | "fill" | "fillred" ->
              let a_in = Var (List.hd iter_inputs) in
              ( a_in::(List.map (array_index i) (List.tl iter_inputs)),
                a_in::(List.map (array_index i) (List.tl iter_outputs)))
            | _ -> assert false (* should not occur *)
          in
          let cont ectx val_ctx = (* necessary? correct? *)
            rval_ctx := val_ctx;
            ref_event := ectx;
            cont ectx val_ctx
          in
          ref_event := do_soc_step lxm inst_name.(i) node_step !rval_ctx soc_tbl
            node_soc vel_in vel_out !ref_event cont;
          
          rval_ctx := { !rval_ctx with cpath = List.tl !rval_ctx.cpath };
        done;
        if iter <> "map" then (
          let a_in  = Var (List.hd iter_inputs) in
          let a_out = Var (List.hd iter_outputs) in
          rval_ctx := assign_expr !rval_ctx a_in a_out);  (* a_out=a_n *)
        cont !ref_event !rval_ctx
    in
    event

and (do_gaol : Soc.tbl -> Event.t -> gao list -> SocExecValue.ctx ->
     (Event.t -> SocExecValue.ctx -> Event.t) ->  Event.t) =
  fun soc_tbl ectx gaol val_ctx cont -> match gaol with
  | [] -> assert false
  | [gao] -> do_gao soc_tbl ectx gao val_ctx cont
  | gao::gaol ->
    let cont ectx val_ctx = do_gaol soc_tbl ectx gaol val_ctx cont in
    do_gao soc_tbl ectx gao val_ctx cont

and (do_gao : Soc.tbl -> Event.t -> gao -> SocExecValue.ctx ->
     (Event.t -> SocExecValue.ctx -> Event.t) ->  Event.t)  =
  fun soc_tbl ectx gao val_ctx cont -> 
    match gao with
    | Case(id, id_gao_l,lxm) -> (
      try 
        let id_val = get_enum id val_ctx in
        let gaol = List.assoc id_val id_gao_l in
        do_gaol soc_tbl ectx gaol val_ctx cont 
      with Not_found -> cont ectx val_ctx 
    )
    | Call(vel_out, Assign, vel_in, lxm) -> (
      let val_ctx = 
        assert (List.length vel_in = List.length vel_out);
        List.fold_left2 assign_expr val_ctx vel_in vel_out
      in
      cont ectx val_ctx 
    )
    | Call(vel_out, Procedure sk, vel_in, lxm) -> (
      let (proc_name,_,_) = sk in
      let path_saved = val_ctx.cpath in
      let val_ctx = { val_ctx with cpath = proc_name::val_ctx.cpath } in
      let soc = SocUtils.find lxm sk soc_tbl in
      let step = match soc.step with [step] -> step | _ ->  assert false in
      let cont ectx val_ctx =
        let val_ctx =  { val_ctx with cpath = path_saved } in
        cont ectx val_ctx
      in
      let cont ectx val_ctx =
        do_soc_step lxm proc_name step val_ctx soc_tbl soc vel_in vel_out ectx cont
      in
      cont ectx val_ctx 
    )
    | Call(vel_out, Method((inst_name,sk),step_name), vel_in, lxm) -> (
      let path_saved = val_ctx.cpath in
      let val_ctx = { val_ctx with cpath = inst_name::val_ctx.cpath } in
      let soc = SocUtils.find lxm sk soc_tbl in
      let step = try List.find (fun sm -> sm.name = step_name) soc.step
        with Not_found -> assert false
      in
      let cont ectx val_ctx =
        let val_ctx = { val_ctx with cpath = path_saved } in
        cont ectx val_ctx
      in
      let cont ectx val_ctx = 
        do_soc_step lxm inst_name step val_ctx soc_tbl soc vel_in vel_out ectx cont
      in
      cont ectx val_ctx 
    )
and (do_soc_step : Lxm.t -> Lv6Id.t -> step_method -> SocExecValue.ctx ->
     Soc.tbl -> Soc.t ->  var_expr list -> var_expr list -> Event.t -> 
     (Event.t -> SocExecValue.ctx -> Event.t) ->  Event.t) =
  fun lxm name step val_ctx soc_tbl soc vel_in vel_out ectx cont -> 
    profile_info ("SocExecDbg.do_soc_step "^name^"\n");
    let soc_in_vars, soc_out_vars = soc.profile in
    let step_in_vars = filter_params soc soc_in_vars step.idx_ins in 
    let step_out_vars = filter_params soc soc_out_vars step.idx_outs in
    let new_s = substitute_args_and_params vel_in step_in_vars val_ctx in
    let val_ctx =  { val_ctx with s = new_s } in
    let cont ectx val_ctx =
      let s_out = substitute_params_and_args step_out_vars vel_out val_ctx in
      cont ectx { val_ctx with s = s_out }
    in
    (*     let (datal:Data.subst list) = get_all_subst val_ctx.s in *)
    let datal = get_input_vals val_ctx step_in_vars in   
    let (soc_name,_,_) = soc.key in
    let step_name = match soc.step with
        [_] -> soc_name
      | _ -> soc_name^"."^step.name
    in
    (*     let soc_step_in, soc_step_out = step.profile in *)
    let sinfo = 
        let l = Lxm.line lxm in
        if l <= 0 then None else
          Some(fun () ->
            let atom =
              {
                Event.str  = Lxm.str lxm; 
                Event.file = Lxm.file lxm ; 
                Event.line = Lxm.line lxm ,Lxm.line lxm ; 
                Event.char = Lxm.cstart lxm, Lxm.cend lxm;
                Event.stack = 
                  match ectx.Event.sinfo with
                  | None -> None
                  | Some si -> Some (List.hd ((si()).Event.atoms));
              }
            in
            {
              Event.expr  = Expr.Var "dummy" ; (* XXX *)
              Event.atoms = [atom];
              Event.more  = None;  (* yet *)
            }
          )
    in
    let initial_sinfo = ectx.sinfo in
    let ectx = {
      ectx with
        Event.name = step_name;
        Event.depth = ectx.Event.depth+1;
        Event.data = datal;
        Event.inputs  = fst soc.profile;
        Event.outputs = snd soc.profile;
        Event.sinfo = sinfo;
    }
    in
    let cont2 local_ectx val_ctx =
(*       let (datal:Data.subst list) = List.map
        (fun v -> fst v, SocExecValue.get_val (fst v) val_ctx) step_out_vars
         in *)
      let (datal:Data.subst list) = SocExecValue.get_vals val_ctx in 
      let nectx = Event.incr_event_nb local_ectx in
      let nectx = Event.decr_event_depth nectx in
      let nectx = { nectx with Event.sinfo = initial_sinfo } in
      {
        Event.step = ectx.Event.step;
        Event.nb = local_ectx.Event.nb;
        Event.depth = ectx.Event.depth;
        Event.kind = Event.Exit;
        Event.lang = "lustre";
        Event.name = step_name;
        Event.inputs = ectx.Event.inputs;
        Event.outputs = ectx.Event.outputs;
        Event.sinfo = sinfo;
        Event.data = datal;
        Event.next = (fun () -> cont nectx val_ctx);
        Event.terminate = ectx.Event.terminate;
      } 
    in
    { 
      Event.step = ectx.Event.step;
      Event.nb = ectx.Event.nb;
      Event.depth = ectx.Event.depth;
      Event.kind = Event.Call;
      Event.lang = "lustre";
      Event.name = step_name;
      Event.inputs = ectx.Event.inputs;
      Event.outputs = ectx.Event.outputs;
      Event.sinfo = sinfo;
      Event.data = ectx.Event.data;
      Event.next = (fun () ->
        assert (List.mem  step soc.step);
        let ectx = Event.incr_event_nb ectx in
        soc_step lxm step soc_tbl soc ectx { val_ctx with s=new_s } cont2
      );
      Event.terminate = ectx.Event.terminate;
    }



(* get the step params from its soc params *)
and (filter_params : Soc.t -> Soc.var list -> int list -> Soc.var list) =
  fun soc el il ->
    let local_nth i l = 
      try List.nth l i 
      with _ -> 
        print_string (
          "\n*** Cannot get the " ^ (string_of_int (i+1)) ^ 
            "the element of a list of size " ^ (string_of_int (List.length l))^"\n");
        flush stdout;
        assert false 
    in
    let res  = List.map (fun i -> local_nth i el) il  in
    res

(* End of XXX duplication of SocExec !  *)
(****************************************************************************)
      
(* exported *) 
and (do_step_dbg : Soc.tbl -> Soc.t -> Event.t -> SocExecValue.ctx ->
     (Event.t -> SocExecValue.ctx -> Event.t) -> Event.t) =
  fun soc_tbl soc ectx val_ctx cont ->
    let soc_in_vars, soc_out_vars = soc.profile in
    (*    let (datal:Data.subst list) = get_all_subst val_ctx.s in  *)
    let (datal:Data.subst list) = get_input_vals val_ctx soc_in_vars in 
    let (soc_name,_,_) = soc.key in
    let lxm = (List.hd soc.step).lxm in
    let sinfo = 
          Some(fun () ->
            let atom =
              {
                Event.str  = soc_name;
                Event.file = Lxm.file lxm; 
                Event.line = Lxm.line lxm,Lxm.line lxm; 
                Event.char = Lxm.cstart lxm, Lxm.cend lxm;
                Event.stack = None;
              }
            in
            {
              Event.expr  = Expr.Var "dummy" ; (* XXX *)
              Event.atoms = [atom];
              Event.more  = None;  (* yet *)
            }
          )
    in
    let ectx = { 
      ectx with 
        Event.name = soc_name;
        Event.depth = ectx.Event.depth+1;
        Event.data = datal;
        Event.inputs  = fst soc.profile;
        Event.outputs = snd soc.profile;
        Event.sinfo = sinfo;
    } 
    in
    let cont2 local_ectx val_ctx =
      (* je pourrais enlver les entrées... *)
      let (datal:Data.subst list) = SocExecValue.get_vals val_ctx in 
      let nectx = Event.incr_event_nb local_ectx in
      let nectx = { ectx with
        Event.nb = local_ectx.Event.nb+1;
        Event.data = datal;
        Event.depth = ectx.Event.depth-1;
      }
      in
      { 
        Event.step = ectx.Event.step;
        Event.nb = local_ectx.Event.nb;
        Event.depth = ectx.Event.depth;
        Event.kind = Event.Exit;
        Event.lang = "lustre";
        Event.name = soc_name;
        Event.inputs  = fst soc.profile;
        Event.outputs = snd soc.profile;
        Event.sinfo = sinfo;
        Event.data = datal;
        Event.next = (fun () -> cont nectx val_ctx);
        Event.terminate = ectx.Event.terminate;
      } 
    in
    { 
      Event.step = ectx.Event.step;
      Event.nb = ectx.Event.nb;
      Event.depth = ectx.Event.depth;
      Event.kind = Event.Call;
      Event.lang = "lustre";
      Event.name = soc_name;
      Event.inputs = ectx.Event.inputs;
      Event.outputs = ectx.Event.outputs;
      Event.sinfo = None;
      Event.data = ectx.Event.data;
      Event.next = (fun () -> 
        let step = match soc.step with [step] -> step | _ ->  assert false in
        let ectx = Event.incr_event_nb ectx in
        soc_step step.lxm step soc_tbl soc ectx val_ctx cont2
      );
      Event.terminate = ectx.Event.terminate;
    }
